from typing import Iterable, Union, OrderedDict

import loralib as lora
import torch
from loguru import logger
from pytorch_lightning import LightningModule
from torch import nn

import tacv.torch


class AnyLoRA(LightningModule):
    """This class is designed to not be used standalone. Use it as an extra parent class in a multiple inheriting class, for example:
    class MyClass(AnyTorchModel, AnyLoRA):
        def __init__(param):
            AnyTorchModel.__init__(param)
    Remember to make the __init__ function the same AnyTorchModel.__init__ function
    """

    def _init_lora_on_layer(self, layer_name: str, rank):
        layer_components = layer_name.split(".")
        parent_component = self
        child_component_name = layer_components[0]
        layer = self.__getattr__(child_component_name)
        if len(layer_components) > 1:
            for component in layer_components[1:]:
                parent_component = layer
                child_component_name = component
                if child_component_name.isdigit():
                    layer = layer[int(child_component_name)]
                else:
                    layer = layer.__getattr__(child_component_name)

        if isinstance(layer, nn.Linear):
            new_layer = lora.Linear(in_features=layer.in_features, out_features=layer.out_features, r=rank)
        elif isinstance(layer, nn.Embedding):
            new_layer = lora.Embedding(num_embeddings=layer.num_embeddings, embedding_dim=layer.embedding_dim,
                                       r=rank)
        elif isinstance(layer, nn.Conv2d):
            new_layer = lora.Conv2d(in_channels=layer.in_channels, out_channels=layer.out_channels,
                                    kernel_size=layer.kernel_size if isinstance(layer.kernel_size, int) else
                                    layer.kernel_size[0], r=rank)
        else:
            logger.warning(
                f"Currently we only support initializing LoRA on layer types [Linear, Conv2d, Embedding]. But found layer name {layer_name} of type {type(layer)}")
            return
        new_layer.weight.data = layer.weight.data
        if layer.bias is not None:
            new_layer.bias.data = layer.bias.data
        parent_component.__setattr__(child_component_name, new_layer)

    def init_lora(self, rank: Union[int, Iterable[int]], layers_start_with: Union[Iterable[str], str] = None,
                  layers_end_with: Union[Iterable[str], str] = None, layers_contain: Union[Iterable[str], str] = None):
        """
        Init loRA weights for layers:
        @param rank: rank of the LoRA weights
        @param layers_start_with: Init LoRA weights on layers having names starting with specified values
        @param layers_end_with: Init LoRA weights on layers having names ending with specified values
        @param layers_contain: Init LoRA weights on layers having names containing specified values
        """
        assert rank > 0 and isinstance(rank, int)
        assert layers_start_with or layers_end_with or layers_contain

        if layers_start_with is None:
            layers_start_with = []
        if layers_end_with is None:
            layers_end_with = []
        if layers_contain is None:
            layers_contain = []

        if isinstance(layers_start_with, str):
            layers_start_with = [layers_start_with]
        if isinstance(layers_end_with, str):
            layers_end_with = [layers_end_with]
        if isinstance(layers_contain, str):
            layers_contain = [layers_contain]

        # get all model param names
        model_param_names = [name for name, _ in self.named_parameters()]
        unique_param_names = set()
        for p_name in model_param_names:
            splits = p_name.split(".")
            if len(splits) > 0:
                if splits[-1] == "weight":
                    unique_param_names.add(".".join(splits[:-1]))
                elif splits[-1] == "bias":
                    unique_param_names.add(".".join(splits[:-1]))
        model_param_names = list(unique_param_names)

        for p_name in model_param_names:
            for layer_contain in layers_contain:
                if layer_contain in p_name:
                    self._init_lora_on_layer(p_name, rank)

        for p_name in model_param_names:
            for layer_end_with in layers_end_with:
                if p_name.endswith(layer_end_with):
                    self._init_lora_on_layer(p_name, rank)

        for p_name in model_param_names:
            for layer_start_with in layers_start_with:
                if p_name.startswith(layer_start_with):
                    self._init_lora_on_layer(p_name, rank)

    def eval(self, break_lora=False):
        if not break_lora:
            logger.warning(
                "Fail to turn on eval mode. once the eval mode is on, we can not back to LoRA-training mode again, use break_lora=True to confirm")
            return
        return super(AnyLoRA, self).eval()

    def save_lora(self, path, bias="none"):
        lora_dict = lora.lora_state_dict(self, bias=bias)
        torch.save(lora_dict, path)

    def load_lora(self, pretrained: Union[str, OrderedDict], map_location="cpu"):
        tacv.torch.load_weights(self, pretrained, location=map_location)

    def mark_only_lora_as_trainable(self, bias="none"):
        lora.mark_only_lora_as_trainable(self, bias)



from typing import Optional, Tuple, Dict, Any, Union, List

import torch
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from pytorch_lightning.loggers import MLFlowLogger
from torch.optim.lr_scheduler import MultiStepLR
from torch.utils.data import DataLoader
from transformers import GPTNeoForCausalLM
from transformers.modeling_outputs import CausalLMOutputWithPast

from .lora_any import AnyLoRA


# Copy from https://github.com/Lightning-AI/lit-gpt/blob/f50e2d8d6903863d6f558b7a1953bb4125ee5f96/lit_gpt/utils.py#L437
# this is to avoid OOM when compute gradient by splitting logits to smaller chunks and compute loss one by one
def chunked_cross_entropy(
    logits: Union[torch.Tensor, List[torch.Tensor]], targets: torch.Tensor, chunk_size: int = 128
) -> torch.Tensor:
    if isinstance(logits, list):
        # don't want to chunk cross entropy
        if chunk_size == 0:
            logits = torch.cat(logits, dim=1)
            logits = logits.reshape(-1, logits.size(-1))
            targets = targets.reshape(-1)
            return torch.nn.functional.cross_entropy(logits, targets, ignore_index=-1)

        # chunk cross entropy
        logit_chunks = [logit_chunk.reshape(-1, logit_chunk.size(-1)) for logit_chunk in logits]
        target_chunks = [target_chunk.reshape(-1) for target_chunk in targets.split(logits[0].size(1), dim=1)]
        loss_chunks = [
            torch.nn.functional.cross_entropy(logit_chunk, target_chunk, ignore_index=-1, reduction="none")
            for logit_chunk, target_chunk in zip(logit_chunks, target_chunks)
        ]
        return torch.cat(loss_chunks).mean()

    # no chunking at all
    logits = logits.reshape(-1, logits.size(-1))
    targets = targets.reshape(-1)
    if chunk_size == 0:
        return torch.nn.functional.cross_entropy(logits, targets, ignore_index=-1)

    # lm_head wasn't chunked, chunk cross entropy
    logit_chunks = logits.split(chunk_size)
    target_chunks = targets.split(chunk_size)
    loss_chunks = [
        torch.nn.functional.cross_entropy(logit_chunk, target_chunk, ignore_index=-1, reduction="none")
        for logit_chunk, target_chunk in zip(logit_chunks, target_chunks)
    ]
    return torch.cat(loss_chunks).mean()

class GPTNeoLoRA(GPTNeoForCausalLM, AnyLoRA):
    def __init__(self, config):
        GPTNeoForCausalLM.__init__(self, config)
        self.lm_head_chunk_size = 128

    # Copy and modified from transformers.GPTNeoForCausalLM
    def forward(
            self,
            input_ids: Optional[torch.Tensor] = None,
            past_key_values: Optional[Tuple[torch.FloatTensor]] = None,
            attention_mask: Optional[torch.Tensor] = None,
            token_type_ids: Optional[torch.Tensor] = None,
            position_ids: Optional[torch.Tensor] = None,
            head_mask: Optional[torch.Tensor] = None,
            inputs_embeds: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None,
            use_cache: Optional[bool] = None,
            output_attentions: Optional[bool] = None,
            output_hidden_states: Optional[bool] = None,
            return_dict: Optional[bool] = None,
            head_chunk_size=-1,
    ):
        r"""
                labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
                    Labels for language modeling. Note that the labels **are shifted** inside the model, i.e. you can set
                    `labels = input_ids` Indices are selected in `[-100, 0, ..., config.vocab_size]` All labels set to `-100`
                    are ignored (masked), the loss is only computed for labels in `[0, ..., config.vocab_size]`
                """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs[0]
        if head_chunk_size > 0:
            # chunk the lm head logits to reduce the peak memory used by autograd
            lm_logits = [self.lm_head(x_i) for x_i in hidden_states.split(head_chunk_size, dim=1)]
        else:
            lm_logits = self.lm_head(hidden_states)

        loss = None
        if not return_dict:
            output = (lm_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutputWithPast(
            loss=loss,
            logits=lm_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
        )

    def configure_optimizers(self):
        self.mark_only_lora_as_trainable(bias="none")
        trainable_params = [p for p in self.parameters() if p.requires_grad]
        num_params = sum(p.numel() for p in trainable_params)
        print(f"Number of trainable parameters: {num_params}")
        num_params = sum(p.numel() for p in self.parameters() if not p.requires_grad)
        print(f"Number of non trainable parameters: {num_params}")

        optimizer = torch.optim.AdamW(trainable_params, lr=0.0002)
        scheduler = MultiStepLR(optimizer=optimizer, milestones=(45, 90, 180), gamma=0.5)
        return [optimizer], [scheduler]

    def training_step(self, batch, batch_id):
        input_ids = batch["x"]
        targets = batch["y"]
        logits = self.forward(input_ids, return_dict=False, head_chunk_size=self.lm_head_chunk_size)[0]

        # shift the targets such that output n predicts token n+1
        logits[-1] = logits[-1][..., :-1, :]
        loss = chunked_cross_entropy(logits, targets[..., 1:])
        self.log("train_loss", loss.item())
        return loss

    def validation_step(self, batch, batch_id):
        input_ids = batch["x"]
        targets = batch["y"]
        logits = self.forward(input_ids, return_dict=False, head_chunk_size=self.lm_head_chunk_size)[0]
        # shift the targets such that output n predicts token n+1
        logits[-1] = logits[-1][..., :-1, :]
        loss = chunked_cross_entropy(logits, targets[..., 1:])
        self.log("val_loss", loss.item())
        return loss

    def on_save_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        state_dict = checkpoint["state_dict"]
        state_dict = {key: val for key, val in state_dict.items() if "lora_" in key}
        checkpoint["state_dict"] = state_dict


# TODO: Define your data loaders
train_data = YourDataset(data_dir / "train.pt")
val_data = YourDataset(data_dir / "test.pt")
val_data.max_seq_length = train_data.max_seq_length

train_loader = DataLoader(train_data, batch_size=8, shuffle=True, pin_memory=True, num_workers=4)
val_loader = DataLoader(val_data, batch_size=8, shuffle=False)

model = GPTNeoLoRA.from_pretrained("EleutherAI/gpt-neo-1.3B", low_cpu_mem_usage=True)
model.init_lora(rank=8, layers_contain=["k_proj", "v_proj"])
model.mark_only_lora_as_trainable(bias="none")

# training callbacks
early_stop = EarlyStopping(monitor="val_loss", patience=10)
model_ckpt = ModelCheckpoint(dirpath="log_dir", filename='lora-epoch{epoch:02d}-val_loss{val_loss:.5f}',
                             monitor="val_loss", save_top_k=4)
logger = MLFlowLogger(experiment_name="lora_test")

trainer = Trainer(accelerator="gpu", devices=2, precision="16-mixed", callbacks=[early_stop, model_ckpt],
                  logger=logger, max_epochs=5000)
trainer.fit(model, train_loader, val_loader)

# Save the final LoRA checkpoint at the end of training
model.save_lora("lora.pth")


"""This code was taken and modified from https://gist.github.com/chandlerprall/1017266"""

import threading
import sys, os, re
import time
from queue import Queue
from urllib.request import urlopen


class ThreadedDownload(object):
    REGEX = {
        'hostname_strip': re.compile('.*\..*?/', re.I)
    }

    class MissingDirectoryException(Exception):
        pass

    class DownloadTracker(threading.Thread):
        def __init__(self, queue):
            super(ThreadedDownload.DownloadTracker, self).__init__()
            self.queue = queue

        def run(self) -> None:
            start_time = time.time()
            queue_start_size = max(1, self.queue.qsize())
            while not self.queue.empty():
                queue_size = self.queue.qsize()
                done_percent = 100 * ((queue_start_size - queue_size) / queue_start_size)
                taken_time = time.time() - start_time
                if done_percent > 0:
                    remaining = (100 - done_percent) * taken_time / done_percent
                    print(f"\rDone {done_percent}%, {int(remaining)} seconds remaining.")
                time.sleep(0.5)

    class Downloader(threading.Thread):
        def __init__(self, queue: Queue, report):
            threading.Thread.__init__(self)
            self.queue = queue
            self.report = report

        def run(self):
            while self.queue.empty() == False:
                url = self.queue.get()

                response = url.download()
                if response == False and url.url_tried < url.url_tries:
                    self.queue.put(url)
                elif response == False and url.url_tried == url.url_tries:
                    self.report['failure'].append(url.url)
                elif response == True:
                    self.report['success'].append(url.url)

                self.queue.task_done()

    class URLTarget(object):
        def __init__(self, url, destination, url_tries):
            self.url = url
            self.destination = destination
            self.url_tries = url_tries
            self.url_tried = 0
            self.success = False
            self.error = None

        def download(self):
            self.url_tried = self.url_tried + 1

            try:
                if os.path.exists(self.destination):  # This file has already been downloaded
                    self.success = True
                    return self.success

                remote_file = urlopen(self.url)
                package = remote_file.read()
                remote_file.close()

                if os.path.exists(os.path.dirname(self.destination)) == False:
                    os.makedirs(os.path.dirname(self.destination))

                dest_file = open(self.destination, 'wb')
                dest_file.write(package)
                dest_file.close()

                self.success = True

            except Exception as e:
                self.error = e

            return self.success

        def __str__(self):
            return 'URLTarget (%(url)s, %(success)s, %(error)s)' % {'url': self.url, 'success': self.success,
                                                                    'error': self.error}

    def __init__(self, urls=[], destination='.', directory_structure=False, thread_count=5, url_tries=3):
        if os.path.exists(destination) is False:
            os.makedirs(destination)

        self.queue = Queue(0)  # Infinite sized queue
        self.report = {'success': [], 'failure': []}
        self.threads = []

        if destination[-1] != os.path.sep:
            destination = destination + os.path.sep
        self.destination = destination
        self.thread_count = thread_count
        self.directory_structure = directory_structure

        # Prepopulate queue with any values we were given
        for url in urls:
            self.queue.put(ThreadedDownload.URLTarget(url, self.fileDestination(url), url_tries))

    def fileDestination(self, url):
        if self.directory_structure == False:
            # No directory structure, just filenames
            file_destination = '%s%s' % (self.destination, os.path.basename(url))

        elif self.directory_structure == True:
            # Strip off hostname, keep all other directories
            file_destination = '%s%s' % (self.destination, ThreadedDownload.REGEX['hostname_strip'].sub('', url))

        elif hasattr(self.directory_structure, '__len__') and len(self.directory_structure) == 2:
            # User supplied a custom regex replace
            regex = self.directory_structure[0]
            if isinstance(regex, str):
                regex = re.compile(str)
            replace = self.directory_structure[1]
            file_destination = '%s%s' % (self.destination, regex.sub(replace, url))

        else:
            # No idea what's wanted
            file_destination = None

        if hasattr(file_destination, 'replace'):
            file_destination = file_destination.replace('/', os.path.sep)
        return file_destination

    def addTarget(self, url, url_tries=3):
        self.queue.put(ThreadedDownload.URLTarget(url, self.fileDestination(url), url_tries))

    def run(self):
        for i in range(self.thread_count):
            thread = ThreadedDownload.Downloader(self.queue, self.report)
            thread.start()
            self.threads.append(thread)
        if self.queue.qsize() > 0:
            download_tracker = ThreadedDownload.DownloadTracker(self.queue)
            download_tracker.run()
            self.queue.join()



import torch
from torch.nn.functional import smooth_l1_loss


def centerness_loss(gt, pred):
    '''
    :param gt:  Bx1xHxW
    :param pred: Bx1xHxW
    :return:
    '''
    pos_inds = gt.eq(1).float()
    neg_inds = gt.lt(1).float()

    neg_weights = torch.pow(1 - gt, 4)
    pos_loss = torch.log(pred) * torch.pow(1 - pred, 2) * pos_inds
    neg_loss = torch.log(1 - pred) * torch.pow(pred, 2) * neg_weights * neg_inds

    num_pos = pos_inds.float().sum()
    pos_loss = pos_loss.sum()
    neg_loss = neg_loss.sum()

    if num_pos == 0:
        loss = - neg_loss
    else:
        loss = - (pos_loss + neg_loss) / num_pos
    return loss


def l1_loss(gt, pred, centermap):
    positives = (centermap == 1).sum(dim=1, keepdim=True)
    loss = positives * smooth_l1_loss(pred, gt,reduction="none")
    return loss.sum()

def non_negative_loss(gt,pred,centermap):
    '''
    This loss function pulls models from predicting cropped boxes. To do that, it penalizes keypoints being
    nearer their centers than its groundtruth counterparts
    :param gt:
    :param pred:
    :param centermap:
    :return:
    '''
    positives = (centermap == 1).sum(dim=1, keepdim=True)
    distance_gt_pred = (pred**2-gt**2)
    zero = torch.zeros_like(distance_gt_pred)
    loss = positives * torch.where(distance_gt_pred>zero,zero,-distance_gt_pred)
    return loss.sum()


from setuptools import setup
import setuptools
import os


def read_version_int(version_file_path):
    if not os.path.exists(version_file_path):
        return None
    with open(version_file_path, "r") as version_file:
        text = version_file.readlines()[0]
        version_file.close()
        batches = text.split(".")
        batches = [int(batch) for batch in batches]
        return batches


def increase_version_by_one(version_file_path):
    batches = read_version_int(version_file_path)
    if batches is None:
        return
    batch1, batch2, batch3 = batches
    batch3 += 1
    if batch3 >= 10:
        batch3 = batch3 % 10
        batch2 += 1
        if batch2 >= 10:
            batch1 = batch1 + 1
    version_file = open(version_file_path, "w+")
    new_ver = str(batch1) + "." + str(batch2) + "." + str(batch3)
    version_file.write(new_ver + "\n")
    version_file.close()
    return new_ver


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open("requirements.txt", "r") as f:
    lines = f.readlines()
    required_pkgs = [item.strip() for item in lines]

# version_file = "tacv/resources/version.txt"
new_ver = "1.1.8"  # increase_version_by_one(version_file)
print(f"Building {new_ver}")
setup(
    name='tacv',
    version=new_ver,
    packages=setuptools.find_packages(where="."),
    url='https://github.com/TaQuangTu/TaCV',
    license='LICENSE',
    author='QuangTu-Ta',
    install_requires=required_pkgs,
    author_email='taquangtu132@gmail.com',
    description='A mini package for daily tasks',
    long_description_content_type="text/markdown",
    long_description=long_description,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
